"""
This module holds some useful functions and classes for the data behavioral package.
"""

from abc import ABCMeta
from typing import Hashable, Iterator, Sequence, TypeVar

from ...utils import chrono

__all__ = ["levenshtein_similarity", "levenshtein_subset_similarity", "entropy", "printable_rate"]





S = TypeVar("S", bound=Sequence[Hashable])

@chrono
def levenshtein_similarity(s1 : S, s2 : S, threshold : float = 1.0) -> float:
    """
    Returns the normalized Levenshtein distance (float between 0 and 1) between two strings, where the weight of insertion and deletion operation is the same than the one of substitution.
    If threshold is given, it must be a float representing the minimum value of similarity required to push the calculation to the end. If similarity gets lower, 0 is returned.
    """
    from typing import Sequence
    if not isinstance(s1, Sequence):
        raise TypeError("Expected two Sequences of Hashables, got " + repr(type(s1).__name__) + " and " + repr(type(s2).__name__))
    if not isinstance(s2, type(s1)):
        raise TypeError("Expected two {}s, got {} and {}".format(type(s1).__name__, repr(type(s1).__name__), repr(type(s2).__name__)))
    if not isinstance(threshold, float):
        try:
            threshold = float(threshold)
        except:
            raise TypeError("Expected float for threshold, got " + repr(type(threshold).__name__)) from None
    if not (0 <= threshold <= 1):
        raise ValueError("Threshold must be between zero and one, got " + repr(threshold))
    if not s1 and not s2:
        return 0.0
    if hash(s1) == hash(s2) and s1 == s2:
        return 1.0
    if threshold >= 1.0:
        score_threshold = None
    else:
        score_threshold = round((1 - threshold) * max(len(s1), len(s2)))
    modifier = 1
    from Levenshtein import distance
    s = 1.0 - (distance(s1, s2, weights = (1, 1, modifier), score_cutoff = score_threshold) / max(len(s1), len(s2)))
    if s <= threshold:
        return 0.0
    return s

@chrono
def levenshtein_subset_similarity(s1 : S, s2 : S, threshold : float = 1.0) -> float:
    """
    Returns the normalized Levenshtein distance (float between 0 and 1) between two strings, where the weight of insertion and deletion operation is lower than the one of substitution.
    If threshold is given, it must be a float representing the minimum value of similarity required to push the calculation to the end. If similarity gets lower, 0 is returned.
    """
    from typing import Sequence
    if not isinstance(s1, Sequence):
        raise TypeError("Expected two Sequences of Hashables, got " + repr(type(s1).__name__) + " and " + repr(type(s2).__name__))
    if not isinstance(s2, type(s1)):
        raise TypeError("Expected two {}s, got {} and {}".format(type(s1).__name__, repr(type(s1).__name__), repr(type(s2).__name__)))
    if not isinstance(threshold, float):
        try:
            threshold = float(threshold)
        except:
            raise TypeError("Expected float for threshold, got " + repr(type(threshold).__name__)) from None
    if not (0 <= threshold <= 1):
        raise ValueError("Threshold must be between zero and one, got " + repr(threshold))
    if not s1 and not s2:
        return 0.0
    if hash(s1) == hash(s2) and s1 == s2:
        return 1.0
    if threshold >= 1.0:
        score_threshold = None
    else:
        score_threshold = round((1 - threshold) * max(len(s1), len(s2)))
    modifier = 3
    from Levenshtein import distance
    s = 1.0 - (distance(s1, s2, weights = (1, 1, modifier), score_cutoff = score_threshold) / max(len(s1), len(s2)))
    if s <= threshold:
        return 0.0
    return s

@chrono
def entropy(s : str | bytes | bytearray) -> float:
    """
    Computes the character entropy of the given string.
    """
    if not s:
        return 0.0
    from collections import Counter
    from math import log2
    probabilities = {c : nc / len(s) for c, nc in Counter(s).items()}
    return sum(-p * log2(p) for p in probabilities.values())

@chrono
def printable_rate(s : bytes | bytearray) -> tuple[float, str]:
    """
    Computes the ratio of printable bytes (which can be decoded to the best endoding). Also returns the best probale encoding.
    """
    if not s:
        return 1.0, "raw"

    import codecs
    n_err = 0

    def count(err : UnicodeError) -> tuple[str, int]:
        """
        Adds one to the decoding errors counter and resume normal decoding.
        """
        nonlocal n_err
        n_err += 1
        if isinstance(err, UnicodeDecodeError | UnicodeTranslateError | UnicodeEncodeError):
            return "", err.start + 1
        else:
            raise ValueError("Unknown UnicodeError")
    
    codecs.register_error("count", count)

    best = float("inf")
    best_enc = ""

    # ("utf-8", "ascii", "utf-16", "utf-32", "gbk")

    for enc in ("ascii", ):
        n_err = 0
        codecs.decode(s, enc, "count")
        if n_err < best:
            best = n_err
            best_enc = enc
    
    return (len(s) - best) / len(s), best_enc if best == 0 else "raw"





class DiffFile:

    """
    These objects are abstract/virtual files in which you can perform IO operations at any place.
    """

    def __init__(self) -> None:
        self.__offsets : list[int] = list()
        self.__arrays : list[bytearray] = list()
        self.__last_dump : bytes | None = None
        self.__computation_index : int = 0
        self.__last_computation_index : int = -1
    
    def arrays(self) -> Iterator[tuple[int, bytes]]:
        """
        Yields all the successives entries in the DiffFile in the form (offset, buffer).
        """
        yield from zip(self.__offsets, self.__arrays)

    @chrono
    def write(self, data : bytes, offset : int):
        """
        Writes a new block of data in the virtual diff file.
        """
        if not isinstance(data, bytes | bytearray | memoryview) or not isinstance(offset, int):
            raise TypeError("Expected readable buffer and int, got " + repr(type(data).__name__) + " and " + repr(type(offset).__name__))
        # First, insert the new block in the right place
        data = bytearray(data)
        for i, (o, a) in enumerate(self.arrays()):
            if offset <= o:
                break
        else:
            i = len(self.__offsets)
        self.__offsets.insert(i, offset)
        self.__arrays.insert(i, data)
        # Second, operate all the possible merges
        end = offset + len(data)
        old_data = bytearray(len(data))
        j = i + 1
        while j < len(self.__offsets):
            o, a = self.__offsets[j], self.__arrays[j]
            if o > end:
                break
            old_data[o - offset : o - offset + len(a)] = a
            j += 1
        old_data[:len(data)] = data
        self.__arrays[i] = old_data
        # Delete useless upcoming blocks
        for k in range(i + 1, j):
            self.__offsets.pop(i + 1)
            self.__arrays.pop(i + 1)
        # Eventually merge with previous block
        if i > 0 and self.__offsets[i - 1] + len(self.__arrays[i - 1]) >= offset:
            self.__offsets.pop(i)
            self.__arrays.pop(i)
            o, a = self.__offsets[i - 1], self.__arrays[i - 1]
            a[offset - o : offset - o + len(old_data)] = old_data
        self.__computation_index += 1
    
    @chrono
    def dump(self) -> bytes:
        """
        Returns a dump of the file content. Any empty space will be represented by '[N bytes]'.
        """
        if self.__last_computation_index != self.__computation_index:
            last_end = self.__offsets[0] if self.__offsets else 0
            d = bytearray()
            for o, a in self.arrays():
                if last_end < o:
                    d.extend("[{} bytes]".format(o - (last_end + len(a))).encode("utf-8"))
                d.extend(a)
                last_end = o + len(a)
            self.__last_computation_index = self.__computation_index
            self.__last_dump = bytes(d)
            return self.__last_dump
        else:
            if self.__last_dump is None:
                raise RuntimeError("Got an empty DiffFile dump, whereas it was marked as updated")
            return self.__last_dump





class IOOperation(metaclass = ABCMeta):

    """
    This class represents a Diff vertex atomic operation in its operation log.
    """

    def __init__(self, offset : int, data : bytes) -> None:
        self.__offset = offset
        self.__data = data
    
    @property
    def start(self) -> int:
        """
        The place in the file where this buffer starts.
        """
        return self.__offset
    
    @property
    def stop(self) -> int:
        """
        The place in the file where this buffer ends.
        """
        return self.__offset + len(self.__data)
    
    @property
    def data(self) -> bytes:
        """
        The content of the buffer.
        """
        return self.__data

    



class Read(IOOperation):

    """
    This is a single read operation in a Diff Vertex log.
    """





class Write(IOOperation):

    """
    This is a single write operation in a Diff Vertex log.
    """





del ABCMeta, Hashable, Iterator, Sequence, TypeVar, chrono, S


"""
This module contains integration protocols for this behavioral package.
"""

from ....logger import logger
from ...build import BuildingPhase
from ...utils import chrono
from ..execution.entities import Call
from ..execution.utils import CallHandler
from . import entities, relations

__all__ = []





logger.info("Loading integrations from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

__existing_keys : dict[str, entities.Key] = {}
__inverted_handles : dict[str, entities.Handle] = {}
__active_handles : dict[entities.Handle, str] = {}
__existing_entries : dict[entities.Key, dict[str, entities.KeyEntry]] = {}
__deleted_entries : dict[entities.Key, dict[str, entities.KeyEntry]] = {}

@chrono
def __create_key_tree(key : str) -> entities.Key:
    """
    Creates all the missing keys in leading to the final key and returns the leaf Key node.
    """
    from ...graph import find_or_create
    from ..network import Host
    from .entities import Key
    from .relations import HasSubKey

    k = None
    last = find_or_create(Host, domain = "host")[0]
    for name in key.replace("\x00", "\uFFFD").split("\\"):
        name = name.lower().title()
        k = None
        for e in last.edges:
            if isinstance(e, HasSubKey) and e.source is last and e.destination.name == name:
                k = e.destination
                break
        if not k:
            k = Key()
            k.name = name
            HasSubKey(last, k)
            __existing_keys[k.path.lower()] = k
        last = k
    
    if not k:
        raise RuntimeError("Trying to create an unnamed registry key!")

    return k

@chrono
def __split_key_and_entry_names(path : str) -> tuple[str, str]:       # Might evolve later
    """
    Splits the given key entry path into the path of the last subkey and the name of the entry.
    """
    i = None
    for j, c in enumerate(reversed(path)):
        if c == "\\":
            i = -j - 1
            break
    if i == None:
        raise ValueError("Expected at least one '\\', got " + repr(path))
    return path[:i], path[i + 1:]

@chrono
def __find_last_key_entry(key : entities.Key, name : str) -> entities.KeyEntry | None:
    """
    Given a registry key and an entry name, returns the last entry known with this name.
    """
    if key in __existing_entries and name in __existing_entries[key]:
        return __existing_entries[key][name]

    



@chrono
def integrate_key_opening(c : Call):
    """
    Creates a new Key node if necessary when a key is opened.
    """
    from ....logger import logger
    from .entities import Handle
    from .relations import HasHandle, UsesKey
    if c.status == 1:
        logger.debug("Opening registry key.")
        if c.arguments.regkey.lower() in __existing_keys:
            k = __existing_keys[c.arguments.regkey.lower()]
        else:
            k = __create_key_tree(c.arguments.regkey)
        h = Handle()
        UsesKey(h, k)
        HasHandle(c.thread.process, h)
        if c.arguments.key_handle.lower() in __inverted_handles:
            logger.warning("Opening an already existing registry key handle.")
            return
        __inverted_handles[c.arguments.key_handle.lower()] = h
        __active_handles[h] = c.arguments.key_handle.lower()
        
@chrono
def integrate_key_enumeration(c : Call):
    """
    Finds the enumerated Key and sub-Key. Creates the sub-Key if it does not exist. Links the sub-Key to the Handle.
    """
    from ....logger import logger
    from .entities import Key
    from .relations import Discovered
    if c.status == 1:
        logger.debug("Enumerating from registry key.")
        k : Key
        sk : Key
        if c.arguments.key_handle.lower() not in __inverted_handles:
            logger.warning("Trying to enumerate a registry key with no known handle.")
            return
        h = __inverted_handles[c.arguments.key_handle.lower()]
        k = h.key
        if c.name != "NtEnumerateKey":
            sk = __create_key_tree(k.path + "\\" + c.arguments.key_name)
        else:
            sk = __create_key_tree(k.path + "\\" + c.arguments.buffer.encode("utf-8").decode("utf-16", errors="replace"))     # Cuckoo tried its best...which is quite bad!
        Discovered(h, sk)

@chrono
def integrate_key_closing(c : Call):
    """
    Closes a Handle associated with a Key.
    """
    from ....logger import logger
    if c.status == 1:
        logger.debug("Closing key handle.")
        if c.arguments.key_handle.lower() not in __inverted_handles:
            logger.warning("Trying to close unseen key handle.")
            return
        h = __inverted_handles.pop(c.arguments.key_handle.lower())
        __active_handles.pop(h)

@chrono 
def integrate_key_deleting(c : Call):
    """
    Marks a registry Key as deleted.
    """
    from ....logger import logger
    from ...config import ColorMap
    if c.status == 1:
        logger.debug("Deleting registry key.")
        if c.arguments.key_handle.lower() not in __inverted_handles:
            logger.warning("Trying to delete key from unseen handle.")
            return
        k = __create_key_tree(c.arguments.regkey)
        k.color = ColorMap.get_config().Deleted
        __existing_keys.pop(k.path.lower())
        if k not in __deleted_entries:
            __deleted_entries[k] = {}
        if k in __existing_entries:
            __deleted_entries[k].update(__existing_entries.pop(k))

@chrono
def integrate_key_value_querying(c : Call):
    """
    Creates a KeyEntry node and marks it as read by thea key Handle.
    """
    from ....logger import logger
    from ...graph import Graph
    from .entities import KeyEntry
    from .relations import ChangesTowards, HasEntry, QueriesEntry
    if c.status == 1:
        logger.debug("Querying key entry.")
        if c.arguments.key_handle.lower() not in __inverted_handles:
            logger.warning("Trying to query value with unseen key handle.")
            return
        
        # Finding the open key handle :
        h = __inverted_handles[c.arguments.key_handle.lower()]
        k = h.key

        # Finding the last key in the path to the entry :
        if c.name == "NtEnumerateValueKey":
            sk_path, name = __split_key_and_entry_names(c.arguments.regkey + "\\" + c.arguments.key_name)
        else:
            sk_path, name = __split_key_and_entry_names(c.arguments.regkey)
        sk = __create_key_tree(sk_path)

        skp = sk
        while skp:
            if skp is k:
                break
            skp = skp.parent_key
            if not skp:
                logger.warning("There might be an unkown symbolic link in the registry or there is a problem:\nQuerying key '{}' from open key '{}'.".format(sk.path, k.path))

        # Creating a new KeyEntry object
        nv = KeyEntry[c.flags.reg_type]()
        nv.process_value(c.arguments.value)
        nv.name = name

        # Looking for an olod value for this entry
        ov = __find_last_key_entry(sk, nv.name)
        if not ov or ov.py_type != nv.py_type or ov.value != nv.value or (k in __deleted_entries and ov.name in __deleted_entries[k] and __deleted_entries[k][ov.name] is ov):    # New value is the first or is different : link to previous if there is one.
            v = nv
            if ov:
                ChangesTowards(ov, nv)
        else:                                                       # Old value is still good : delete new value.
            v = ov
            for g in Graph.active_graphs():
                g.remove(nv)
        
        if k not in __existing_entries:
            __existing_entries[k] = {}
        __existing_entries[k][v.name] = v
        if v not in sk.neighbors():
            HasEntry(sk, v)
        QueriesEntry(v, h)

@chrono
def integrate_key_value_setting(c : Call):
    """
    Creates a KeyEntry node and marks it as written by a key Handle.
    """
    from ....logger import logger
    from ...graph import Graph
    from .entities import KeyEntry
    from .relations import ChangesTowards, HasEntry, SetsEntry
    if c.status == 1:
        logger.debug("Setting key entry.")
        if c.arguments.key_handle.lower() not in __inverted_handles:
            logger.warning("Trying to set value with unseen key handle.")
            return
        
        # Finding the open key handle :
        h = __inverted_handles[c.arguments.key_handle.lower()]
        k = h.key

        # Finding the last key in the path to the entry :
        sk_path, name = __split_key_and_entry_names(c.arguments.regkey)
        sk = __create_key_tree(sk_path)

        skp = sk
        while skp:
            if skp is k:
                break
            skp = skp.parent_key
            if not skp:
                logger.warning("There might be an unkown symbolic link in the registry or there is a problem:\nSetting key '{}' from open key '{}'.".format(sk.path, k.path))

        # Creating a new KeyEntry object
        nv = KeyEntry[c.flags.reg_type]()
        nv.process_value(c.arguments.value)
        nv.name = name

        # Looking for an olod value for this entry
        ov = __find_last_key_entry(sk, nv.name)
        if not ov or ov.py_type != nv.py_type or ov.value != nv.value or (k in __deleted_entries and ov.name in __deleted_entries[k] and __deleted_entries[k][ov.name] is ov):    # New value is the first or is different : link to previous if there is one.
            v = nv
            if ov:
                ChangesTowards(ov, nv)
        else:                                                       # Old value is still good : delete new value.
            v = ov
            for g in Graph.active_graphs():
                g.remove(nv)
        
        if k not in __existing_entries:
            __existing_entries[k] = {}
        __existing_entries[k][v.name] = v
        if v not in sk.neighbors():
            HasEntry(sk, v)
        SetsEntry(h, v)

@chrono
def integrate_key_value_deleting(c : Call):
    """
    Creates a KeyEntry node and marks it as deleted by a key Handle.
    """
    from ....logger import logger
    from ...config import ColorMap
    from ...graph import Graph
    from .entities import KeyEntry
    from .relations import ChangesTowards, DeletesEntry, HasEntry
    if c.status == 1:
        logger.debug("Deleting key entry.")
        if c.arguments.key_handle.lower() not in __inverted_handles:
            logger.warning("Trying to set value with unseen key handle.")
            return
        
        # Finding the open key handle :
        h = __inverted_handles[c.arguments.key_handle.lower()]
        k = h.key

        # Finding the last key in the path to the entry :
        sk_path, name = __split_key_and_entry_names(c.arguments.regkey)
        sk = __create_key_tree(sk_path)

        skp = sk
        while skp:
            if skp is k:
                break
            skp = skp.parent_key
            if not skp:
                logger.warning("There might be an unkown symbolic link in the registry or there is a problem:\nDeleting key '{}' from open key '{}'.".format(sk.path, k.path))

        # Creating a new KeyEntry object

        # Looking for an old value for this entry
        ov = __find_last_key_entry(sk, name)
        if not ov:                                                  # New value is the first.
            raise RuntimeError("Here is what you got " + repr(c))
        else:                                                       # Old value : link to new value and delete new.
            nv = KeyEntry[ov.reg_type]()                            # type: ignore      Indeed the type-checker is too dump to know about __class_getitem__...
            ChangesTowards(ov, nv)
            v = nv
            nv.process_value(ov.value)
            for g in Graph.active_graphs():
                g.remove(nv)
        
        if k not in __existing_entries:
            __existing_entries[k] = {}
        __existing_entries[k][v.name] = v
        if k not in __deleted_entries:
            __deleted_entries[k] = {}
        __deleted_entries[k][v.name] = v
        v.color = ColorMap.get_config().Deleted
        if v not in sk.neighbors():
            HasEntry(sk, v)
        DeletesEntry(h, v)





# Key opening
CallHandler(integrate_key_opening, "RegOpenKeyExA", "NtOpenKeyEx", "NtOpenKey", "RegOpenKeyExW")

# Key creation
CallHandler(integrate_key_opening, "NtCreateKey", "RegCreateKeyExW", "RegCreateKeyExA")

# Key enumeration
CallHandler(integrate_key_enumeration, 'RegEnumKeyW', 'NtEnumerateKey', 'RegEnumKeyExW', 'RegEnumKeyExA')

# Key closing
CallHandler(integrate_key_closing, 'RegCloseKey')

# Key deleting
CallHandler(integrate_key_deleting, 'NtDeleteKey', 'RegDeleteKeyW', 'RegDeleteKeyA')

# Key entry querying
CallHandler(integrate_key_value_querying, 'NtQueryValueKey', 'RegQueryValueExW', 'RegQueryValueExA', "RegEnumValueW", 'RegEnumValueA', 'NtEnumerateValueKey')

# Key entry setting
CallHandler(integrate_key_value_setting, 'RegSetValueExW', 'RegSetValueExA', 'NtSetValueKey')

# Key entry deleting
CallHandler(integrate_key_value_deleting, 'RegDeleteValueW', 'RegDeleteValueA', 'NtDeleteValueKey')





__N_referencing_phase = BuildingPhase.request_finalizing_phase()

def find_filesystem_references(ev : BuildingPhase):
    """
    When called with the right finalizing phase event, will cause all KeyEntry nodes to link to File or Folder nodes that their value reference if these entries were modified.
    """
    from ....logger import logger
    from ...utils import is_path, parse_command_line, path_factory
    from ..filesystem import Contains, Directory, File, HasDrive
    from ..network import Host
    from .entities import KeyEntry, Key_SZ_Entry, Key_MULTI_SZ_Entry
    from .relations import DeletesEntry, ReferencesFileSystem, SetsEntry
    if ev.major == "Finalizer" and ev.minor == __N_referencing_phase:
        logger.debug("Finding references to filesystem in {} registry key entries.".format(len(KeyEntry)))
        for ke in KeyEntry:

            if not isinstance(ke, Key_SZ_Entry | Key_MULTI_SZ_Entry):
                continue

            ok = False

            for e in ke.edges:
                if isinstance(e, SetsEntry | DeletesEntry):
                    ok = True
                    break
            if not ok:
                continue

            ok = False
            
            if isinstance(ke, Key_SZ_Entry):
                base_possibilities : list[str] = [ke.value]
            else:
                base_possibilities : list[str] = ke.value

            possibilities = []
            for p in base_possibilities:
                try:
                    args = parse_command_line(p)
                    possibilities.extend(args)
                except KeyboardInterrupt:
                    raise
                finally:
                    possibilities.append(p)
            
            for p in possibilities:
                if is_path(p) and p:
                    p = path_factory(p)
                    current : Host | Directory | File = Host.current
                    next : Directory | File | None
                    parts = list(p.parts)
                    if ":" in parts[0]:
                        parts[0] = parts[0].replace("\\", "")
                    for pi in parts:
                        next = None
                        for e in current.edges:
                            if isinstance(e, HasDrive | Contains) and e.source is current and e.destination.name.lower() == pi.lower():
                                next = e.destination
                                break
                        if next is None:
                            break
                        current = next
                    else:
                        ReferencesFileSystem(ke, current)
        
        logger.debug("Got {} references to filesystem nodes.".format(len(ReferencesFileSystem)))

BuildingPhase.add_callback(find_filesystem_references)





del BuildingPhase, Call, CallHandler, chrono, entities, find_filesystem_references, integrate_key_closing, integrate_key_deleting, integrate_key_enumeration, integrate_key_opening, integrate_key_value_deleting, integrate_key_value_querying, integrate_key_value_setting, logger, relations


"""
This module defines some useful functions for manipulating the BAGUETTE graph typing system.
"""

from functools import cache
from types import ModuleType
from .. import types as type_package
from ..graph import Edge, Vertex

__all__ = ["types", "relations", "relation_types"]





@cache
def types(mod : ModuleType = type_package) -> set[type[Vertex]]:
    """
    Returns the list of available Vertex types available in a given module or package (recursively). Defaults to the BAGUETTE's types package.
    """
    from ..graph import Vertex
    from types import ModuleType

    if not isinstance(mod, ModuleType):
        raise TypeError("Expected module or package, got " + repr(type(mod).__name__))

    type_set = set()
    package_name = mod.__name__

    for name in dir(mod):
        value = getattr(mod, name)
        if isinstance(value, type) and issubclass(value, Vertex):
            type_set.add(value)
        elif isinstance(value, ModuleType):
            name = value.__name__.rpartition(".")[0]
            if name == package_name:
                type_set.update(types(value))
    
    return type_set





@cache
def relations(mod : ModuleType = type_package) -> set[type[Edge]]:
    """
    Returns the list of available Edge and Arrow types available in a given module or package (recursively). Defaults to the BAGUETTE's types package.
    """
    from ..graph import Edge
    from types import ModuleType

    if not isinstance(mod, ModuleType):
        raise TypeError("Expected module or package, got " + repr(type(mod).__name__))

    type_set = set()
    package_name = mod.__name__

    for name in dir(mod):
        value = getattr(mod, name)
        if isinstance(value, type) and issubclass(value, Edge):
            type_set.add(value)
        elif isinstance(value, ModuleType):
            name = value.__name__.rpartition(".")[0]
            if name == package_name:
                type_set.update(types(value))
    
    return type_set





@cache
def relation_types(edge_class : type[Edge]) -> tuple[type[Vertex], type[Vertex]]:
    """
    Given an Edge or Arrow subclass, gives the best type hints for the source and destination vertices.
    Defaults to (Vertex, Vertex) when no annotations exist.
    """
    from typing import get_type_hints
    from ..graph import Edge, Vertex

    if not isinstance(edge_class, type) or not issubclass(edge_class, Edge):
        raise TypeError("Expected Edge subclass, got " + repr(type(edge_class).__name__))

    hints = get_type_hints(edge_class)
    return hints.get("source", Vertex), hints.get("destination", Vertex)
    




del cache, ModuleType, type_package, Edge, Vertex


"""
This module contains integration protocols for this behavioral package.
"""

from typing import Type

from ....logger import logger
from ...build import BuildingPhase
from ...event import Event
from ...graph import Arrow, Edge
from ...utils import chrono
from ..execution.entities import Call
from ..execution.utils import CallHandler
from . import entities, relations

__all__ = ["NewFile", "declare_existing_file"]





logger.info("Loading integrations from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

class NewFile(Event):

    """
    This is an Event thrown when a File Vertex is created.
    """

    __slots__ = {
        "file" : "The File Vertex of the file that was created"
    }

    def __init__(self, file : entities.File) -> None:
        self.file = file





namespace = {name : getattr(relations, name) for name in dir(relations)}
__file_to_dir_cls_table : dict[Type[Arrow | Edge], Type[Arrow | Edge]] = {}
for name, cls in filter(lambda c : isinstance(c[1], type) and issubclass(c[1], Arrow), namespace.items()):
    if name.endswith("File") and name[:-4] + "Directory" in namespace:
        __file_to_dir_cls_table[cls] = namespace[name[:-4] + "Directory"]
    del name, cls

__active_handles : dict[entities.Handle, str] = {}
__inverted_handles : dict[str, entities.Handle] = {}
__existing_files : dict[str, entities.File] = {}
__all_files : dict[str, list[entities.File]] = {}

def declare_existing_file(path : str):
    """
    Declares an existing file in the execution environment.
    """
    from ...utils import is_path
    from .entities import File
    if not isinstance(path, str):
        raise TypeError("Expected str, got " + repr(type(path).__name__))
    if not is_path(path):
        raise ValueError("Expected a valid file path for the execution platform, got " + repr(path))
    f = File()
    f.path = path
    if path not in __all_files:
        __all_files[path] = []
    __all_files[path].append(f)
    __existing_files[path] = f
    NewFile(f).throw()
    




@chrono
def integrate_file_creation(c : Call):
    """
    Creates a File/Directory vertex if it does not exists.
    """
    from ....logger import logger
    from .entities import File, Handle
    from .relations import CreatesFile, HasHandle, UsesFile
    if c.status == 1:
        logger.debug("File created.")
        if c.arguments.file_handle in __inverted_handles:
            h = __inverted_handles[c.arguments.file_handle]
        else:
            h = Handle()
            HasHandle(c.thread.process, h)
        if c.arguments.filepath in __existing_files:
            f = __existing_files[c.arguments.filepath]
            existed = True
        else:
            f = File()
            f.path = c.arguments.filepath
            existed = False
        CreatesFile(c, h)
        if not f in set(h.neighbors()):
            UsesFile(h, f)
        __inverted_handles[c.arguments.file_handle] = h
        __active_handles[h] = c.arguments.file_handle
        __existing_files[c.arguments.filepath] = f
        if c.arguments.filepath not in __all_files:
            __all_files[c.arguments.filepath] = []
        if not existed:
            __all_files[c.arguments.filepath].append(f)
        
        NewFile(f).throw()
        # if "FILE_LIST_DIRECTORY"
        # if "DELETE"
        # if "WRITE_DAC"
        # if "FILE_WRITE_ATTRIBUTES"
        # if "READ_CONTROL"
        # if "FILE_READ_EA"
        # if "FILE_WRITE_DATA"
        # if "FILE_WRITE_EA"
        # if "FILE_APPEND_DATA"
        # if "FILE_READ_ATTRIBUTES"
        # if "GENERIC_WRITE"
        # if "SYNCHRONIZE"
        # if "GENERIC_ALL"
        # if "FILE_READ_DATA" in c.flags.desired_access.split("|"):
        #     h.read = True
        # if "FILE_READ"

@chrono
def integrate_file_opening(c : Call):
    """
    Creates an File/Directory vertex.
    """
    from ....logger import logger
    from .entities import File, Handle
    from .relations import HasHandle, Opens, UsesFile
    if c.status == 1:
        logger.debug("File opened.")
        if c.arguments.file_handle in __inverted_handles:
            h = __inverted_handles[c.arguments.file_handle]
        else:
            h = Handle()
            HasHandle(c.thread.process, h)
        if c.arguments.filepath in __existing_files:
            f = __existing_files[c.arguments.filepath]
            existed = True
        else:
            f = File()
            f.path = c.arguments.filepath
            existed = False
        Opens(c, h)
        if not f in set(h.neighbors()):
            UsesFile(h, f)
        __inverted_handles[c.arguments.file_handle] = h
        __active_handles[h] = c.arguments.file_handle
        __existing_files[c.arguments.filepath] = f
        if c.arguments.filepath not in __all_files:
            __all_files[c.arguments.filepath] = []
        if not existed:
            __all_files[c.arguments.filepath].append(f)

        NewFile(f).throw()

@chrono
def integrate_file_closing(c : Call):
    """
    Connects a closing system call to the corresponding Handle.
    """
    from ....logger import logger
    from .relations import Closes
    if c.status == 1 and c.arguments.handle in __inverted_handles:
        logger.debug("File closed.")
        h = __inverted_handles.pop(c.arguments.handle)
        __active_handles.pop(h)
        Closes(c, h)

@chrono
def integrate_file_reading(c : Call):
    """
    Creates a Data vertex, and links it to the Handle and Call vertices.
    """
    from ....logger import logger
    from ..data import Data
    from ..data.integration import register_read_operation
    from .entities import File
    from .relations import Conveys, Reads
    if c.status == 1 and c.arguments.file_handle in __inverted_handles:
        logger.debug("Reading from file.")
        h = __inverted_handles[c.arguments.file_handle]
        d = Data()
        d.time = c.time
        d.set_data(c.arguments.buffer)
        f = h.file
        Reads(d, c)
        Conveys(h, d)
        if isinstance(f, File):
            register_read_operation(f, h, d.data, c.arguments.offset)

@chrono
def integrate_file_writing(c : Call):
    """
    Creates a Data vertex, and links it to the Handle and Call vertices.
    """
    from ....logger import logger
    from ..data import Data
    from ..data.integration import register_write_operation
    from .entities import File
    from .relations import Conveys, Writes
    if c.status == 1 and c.arguments.file_handle in __inverted_handles:
        logger.debug("Writing to file.")
        h = __inverted_handles[c.arguments.file_handle]
        d = Data()
        d.time = c.time
        d.set_data(c.arguments.buffer)
        f = h.file
        Writes(c, d)
        Conveys(d, h)
        if isinstance(f, File):
            register_write_operation(f, h, d.data, c.arguments.offset)
    
@chrono
def integrate_file_copying(c : Call):
    """
    Creates a File vertex, and links it to the original file it was copied from.
    """
    from ....logger import logger
    from .entities import File
    from .relations import IsCopyiedInto
    if c.status == 1:
        logger.debug("Copying file.")
        if c.arguments.oldfilepath in __existing_files:
            fs = __existing_files[c.arguments.oldfilepath]
        else:
            fs = File()
            fs.path = c.arguments.oldfilepath
            __existing_files[c.arguments.oldfilepath] = fs
            if c.arguments.oldfilepath not in __all_files:
                __all_files[c.arguments.oldfilepath] = []
            __all_files[c.arguments.oldfilepath].append(fs)
            NewFile(fs).throw()
        if c.arguments.newfilepath in __existing_files:
            fd = __existing_files[c.arguments.newfilepath]
        else:
            fd = File()
            fd.path = c.arguments.newfilepath
            __existing_files[c.arguments.newfilepath] = fd
            if c.arguments.newfilepath not in __all_files:
                __all_files[c.arguments.newfilepath] = []
            __all_files[c.arguments.newfilepath].append(fd)
            NewFile(fd).throw()
        IsCopyiedInto(fs, fd)

@chrono
def integrate_file_deleting(c : Call):
    # You need to make it possible to have multiple times the same file name...
    from ....logger import logger
    from ...config import ColorMap
    from .entities import File
    if c.status == 1:
        logger.debug("Deleting file.")
        if c.arguments.filepath not in __existing_files:
            f = File()
            f.path = c.arguments.filepath
            __existing_files[c.arguments.filepath] = f
            if c.arguments.filepath not in __all_files:
                __all_files[c.arguments.filepath] = []
            __all_files[c.arguments.filepath].append(f)
            NewFile(f).throw()
        else:
            f = __existing_files[c.arguments.filepath]
        f.color = ColorMap.get_config().Deleted
        __existing_files.pop(c.arguments.filepath)





# File creation
CallHandler(integrate_file_creation, "NtCreateFile")

# File closing
CallHandler(integrate_file_opening, "NtOpenFile")

# File closing
CallHandler(integrate_file_closing, "NtClose")

# File reading
CallHandler(integrate_file_reading, "NtReadFile")

# File writing
CallHandler(integrate_file_writing, "NtWriteFile")

# File copying
CallHandler(integrate_file_copying, "CopyFileA", "CopyFileExW", "CopyFileW")

# File deleting
CallHandler(integrate_file_deleting, "DeleteFileW", "NtDeleteFile")





__N_phase = BuildingPhase.request_finalizing_phase()

def build_fs_tree(e : BuildingPhase):
    """
    When all File nodes have been created, this will replace those which are actually directories into Directory nodes and build the file system tree.
    """
    # TODO : You will need to make a version of this which keeps track of when files/directories exists to make sure your file system tree is indeed a tree !!!
    from pathlib import PurePath
    from typing import Type

    from ....logger import logger
    from ...graph import Edge, Graph, find_or_create
    from ..execution import Runs
    from ..network import Host
    from .entities import Directory, File
    from .relations import Contains, HasDrive

    disappearing : set[Type[Edge]] = {Runs}


    def mutate(f : File):
        d = Directory()
        d.path = str(f.path)
        d.size = f.size
        graphs = Graph.active_graphs()
        for e in f.edges.copy():
            src, dst = e.source if e.source is not f else d, e.destination if e.destination is not f else d
            e.delete()
            for g in graphs:
                g.remove(e)
            if type(e) in __file_to_dir_cls_table:
                T = __file_to_dir_cls_table[type(e)]
            else:
                T = type(e)
            if T not in disappearing:
                T(src, dst)
        for g in graphs:
            g.remove(f)
            g.append(d)
        return d
    

    if e.major == "Finalizer" and e.minor == __N_phase:
        logger.debug("Mutating up to {} File nodes.".format(len(File)))

        directories : dict[PurePath, list[Directory]] = {}
        files : dict[PurePath, list[File]] = {}
        for filelist in __all_files.values():
            path = filelist[0].path
            for file in filelist:
                paths = [path] + list(path.parents)
                p = paths.pop(0)
                if p not in files:
                    files[p] = []
                files[p].append(file)
                for dpath in paths:
                    if dpath not in directories:
                        directories[dpath] = []
                
        n = 0
        for path in files.copy():
            if path in directories:
                l = list(mutate(file) for file in files.pop(path))
                directories[path].extend(l)
                n += len(l)
        logger.debug("Actually mutated {} File nodes into Directory Nodes.".format(n))
    
        logger.debug("Creating missing directory nodes.")
        for path, filelist in directories.items():
            if not filelist:
                d = Directory()
                d.path = str(path)
                filelist.append(d)
            
        logger.debug("Building file system graph.")
        work : set[File | Directory] = {f for filelist in files.values() for f in filelist}

        while work:
            node = work.pop()
            path = node.path
            parent = path.parent
            if parent != path:
                for parent_node in directories[parent]:
                    Contains(parent_node, node)
                    work.add(parent_node)
            else:
                HasDrive(find_or_create(Host, domain = "host")[0], node)
        
        # Deleting references to irrelevent (mutated) File nodes
        __existing_files.clear()
        __all_files.clear()
                




# TODO : Add a backtracking phase for actual successive copies of files (those without modifications in between)!!!

BuildingPhase.add_callback(build_fs_tree)





del Arrow, BuildingPhase, Call, CallHandler, Edge, Event, Type, build_fs_tree, chrono, entities, integrate_file_closing, integrate_file_copying, integrate_file_creation, integrate_file_deleting, integrate_file_opening, integrate_file_reading, integrate_file_writing, logger, namespace, relations


"""
This module defines some useful tools for MetaGraphs.
"""

from types import UnionType

__all__ = ["class_union"]





def class_union(*cls : type) -> type | UnionType:
    """
    Creates the union of classes, as sum operator but for the "|" operator.
    """
    for c in cls:
        if not isinstance(c, type):
            raise TypeError("Expected types, got " + repr(type(c).__name__))
    l = list(cls)
    if not l:
        raise ValueError("Expected at least one type.")
    c = l.pop()
    while l:
        c = c | l.pop()
    return c





del UnionType


"""
This package defines the graph-related subclasses that modelize the flow of data.
"""

from ....logger import logger





logger.info("Finalizing {} library.".format(__name__.rpartition(".")[2]))

from .entities import *
from .relations import *


"""
This module contains Vertex subclasses for this behavioral package.
"""

from abc import abstractmethod
from types import NoneType
from typing import Any, Callable, Generic, Optional, TypeVar

from ....logger import logger
from ...colors import Color
from ...graph import UniqueVertex, Vertex

__all__ = ["Key", "KeyEntry", "Handle"]





logger.info("Loading entities from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

class Key(UniqueVertex):

    """
    A registry key vertex. Represents a key without its associated value(s).
    """

    __slots__ = {
        "__name" : "The local name of the registry key int the registry tree."
    }

    __pickle_slots__ = {
        "name"
    }

    def __init__(self, *, c: Color | None = None, parent: Vertex | None = None) -> None:
        super().__init__(c=c, parent=parent)
        self.__name : str = ""

    @property
    def name(self) -> str:
        """
        The (local) name of the registry key.
        """
        return self.__name
    
    @name.setter
    def name(self, value : str):
        if not isinstance(value, str):
            raise TypeError("Expected str, got " + repr(type(value).__name__))
        if "\\" in value:
            raise ValueError("'\\' not permitted in a registry key name")
        if len(value) > 255:
            raise ValueError("Key name is too long to be allowed by registry")
        self.__name = value.lower().title().replace("\x00", "\uFFFD")
        if not self.printable:
            from ....logger import logger
            from ...config import ColorMap
            logger.info("Got an unprintable registry key.")
            self.color = ColorMap.get_config().registry.UnprintableKey
    
    @property
    def path(self) -> str:
        """
        The global key name (path from root key and all sub-keys to reach this key included).
        """
        if not self.parent_key:
            return self.name
        return self.parent_key.path + "\\" + self.name

    @property
    def printable(self) -> bool:
        """
        True if the (relative) key name is printable.
        """
        return self.name.isprintable() and "\uFFFD" not in self.name

    @property
    def parent_key(self) -> Optional["Key"]:
        """
        The parent key of this registry key if any (None for root keys).
        """
        from .relations import HasSubKey
        for e in self.edges:
            if isinstance(e, HasSubKey) and e.destination is self and isinstance(e.source, Key):
                return e.source
    
    @property
    def entries(self) -> list["KeyEntry"]:
        """
        All the known entries of this key.
        """
        from .relations import HasEntry
        return [e.destination for e in self.edges if isinstance(e, HasEntry)]
    
    @property
    def label(self) -> str:
        """
        The label of this node.
        """
        return "Key " + self.name





T = TypeVar("T")

class KeyEntry(UniqueVertex, Generic[T]):

    """
    A registry key entry vertex. This represents a value (with name, type and actual value) that a registry key has.
    """

    __slots__ = {
        "__name" : "The name of the registry key entry",
        "__value" : "The actual value stored in this entry"
    }

    __pickle_slots__ = {
        "name",
        "value"
    }

    def __init__(self, *, c: Color | None = None, parent: Optional["Vertex"] = None) -> None:
        super().__init__(c=c, parent=parent)
        self.__name : str = ""

    @property
    def name(self) -> str:
        """
        The name of this registry key entry.
        """
        return self.__name
    
    @name.setter
    def name(self, value : str):
        if not isinstance(value, str):
            raise TypeError("Expected str, got " + repr(type(value).__name__))
        if len(value) > 16383:
            raise ValueError("Registry key value name is too long.")
        self.__name = value
    
    py_type : type[T] | None = None
    reg_type : str | None = None

    __types : dict[str, type["KeyEntry"]] = {}

    def __init_subclass__(cls) -> None:
        res = super().__init_subclass__()
        from types import GenericAlias, UnionType
        if not isinstance(cls.py_type, type | GenericAlias | UnionType) or not isinstance(cls.reg_type, str):
            raise ValueError("Cannot subclass KeyEntry without setting 'py_type' and 'reg_type' to type and str values.")
        if cls.reg_type in KeyEntry.__types:
            raise KeyError("A subclass of KeyEntry for reg_type '{}' already exists.".format(cls.reg_type))
        KeyEntry.__types[cls.reg_type] = cls
        return res
    
    @staticmethod
    def key_types() -> list[str]:
        """
        Lists all registered KeyEntry subclasses names (the native name of the entry type they represent).
        Use 'KeyEntry[<name>]' to get one of the corresponding subclasses.
        """
        return list(KeyEntry.__types)

    def __class_getitem__(cls, key : type | str):
        if isinstance(key, str) and key in KeyEntry.__types:
            return KeyEntry.__types[key]
        return super().__class_getitem__(key) # type: ignore
                
    @property
    def value(self) -> T:
        """
        The value of the entry.
        """
        return self.__value

    @value.setter
    @abstractmethod
    def value(self, val : T):
        raise NotImplementedError

    def _set_value(self, val : T):
        """
        Internal method to change the value in the base class.
        """
        self.__value = val
    
    @abstractmethod
    def process_value(self, val : str | int) -> None:
        """
        Processes the raw value from an API node and sets the result as this entry's value.
        """
        raise NotImplementedError
                    
    @property
    def key(self) -> Key:
        """
        The registry Key that this entry is part of.
        """
        from .relations import HasEntry
        for e in self.edges:
            if isinstance(e, HasEntry):
                return e.source
        raise RuntimeError("Found a KeyEntry that has not been affected to a Key")
    
    @property
    def label(self) -> str:
        return "KeyEntry"
    




class Key_DWORD_Entry(KeyEntry[int]):

    py_type = int
    reg_type = "REG_DWORD"
    
    @property
    def value(self) -> int:
        return super().value
    
    @value.setter
    def value(self, val : int):
        if not isinstance(val, int):
            raise TypeError("Expected int, got " + repr(type(val).__name__))
        self._set_value(val)

    def process_value(self, val: int):
        if not isinstance(val, int):
            raise TypeError("Expected int, got " + repr(type(val).__name__))
        self.value = val

class Key_QWORD_LITTLE_ENDIAN_Entry(Key_DWORD_Entry):
    reg_type = "REG_QWORD_LITTLE_ENDIAN"
    
class Key_DWORD_LITTLE_ENDIAN_Entry(Key_DWORD_Entry):
    reg_type = "REG_DWORD_LITTLE_ENDIAN"
    
class Key_DWORD_BIG_ENDIAN_Entry(Key_DWORD_Entry):
    reg_type = "REG_DWORD_BIG_ENDIAN"
    
class Key_QWORD_Entry(Key_DWORD_Entry):
    reg_type = "REG_QWORD"
    




class Key_SZ_Entry(KeyEntry[str]):

    py_type = str
    reg_type = "REG_SZ"
    
    @property
    def value(self) -> str:
        return super().value
    
    @value.setter
    def value(self, val : str):
        if not isinstance(val, str):
            raise TypeError("Expected str, got " + repr(type(val).__name__))
        self._set_value(val)

    def process_value(self, val: str):
        if not isinstance(val, str):
            raise TypeError("Expected str, got " + repr(type(val).__name__))
        self.value = val

class Key_EXPAND_SZ_Entry(Key_SZ_Entry):
    reg_type = "REG_EXPAND_SZ"
    
class Key_LINK_Entry(Key_SZ_Entry):
    reg_type = "REG_LINK"





class Key_MULTI_SZ_Entry(KeyEntry[list[str]]):

    py_type = list[str]
    reg_type = "REG_MULTI_SZ"
    
    @property
    def value(self) -> list[str]:
        return super().value
    
    @value.setter
    def value(self, val : list[str]):
        if not isinstance(val, list) or any(not isinstance(vi, str) for vi in val):
            raise TypeError("Expected list of str, got " + repr(type(val).__name__))
        self._set_value(val)

    def process_value(self, val: str):
        if not isinstance(val, str):
            raise TypeError("Expected str, got " + repr(type(val).__name__))
        if val.endswith("\x00"):
            v = val[:-1]
        else:
            v = val
        self.value = v.split("\x00")





class Key_BINARY_Entry(KeyEntry[bytes]):

    py_type = bytes
    reg_type = "REG_BINARY"
    
    @property
    def value(self) -> bytes:
        return super().value
    
    @value.setter
    def value(self, val : bytes):
        if not isinstance(val, bytes):
            raise TypeError("Expected bytes, got " + repr(type(val).__name__))
        self._set_value(val)

    def process_value(self, val: str):
        if not isinstance(val, str):
            raise TypeError("Expected str, got " + repr(type(val).__name__))
        self.value = val.encode()





class Key_NONE_Entry(KeyEntry[NoneType]):

    py_type = type(None)
    reg_type = "REG_NONE"
    
    @property
    def value(self) -> NoneType:
        return None
    
    @value.setter
    def value(self, val : NoneType):
        raise ValueError("Cannot set value of REG_NONE registry key entry")
    
    def process_value(self, val: str):
        pass
    




class Handle(UniqueVertex):

    """
    A handle vertex. Represents a registry key handle, used when a program opens a registry key.
    """

    def __init__(self, *, c: Color | None = None, parent: Optional[Vertex] = None) -> None:
        super().__init__(c=c, parent=parent)
    
    @property
    def key(self) -> Key:
        """
        Returns the key node that this handle is working on.
        """
        from .relations import UsesKey
        for e in self.edges:
            if isinstance(e, UsesKey):
                return e.destination
        raise RuntimeError("Key handle with no attached key")
    




del Any, Color, Generic, Optional, TypeVar, UniqueVertex, Vertex, logger, T, Callable, NoneType, abstractmethod


"""
This module contains Edge and Arrow subclasses for this behavioral package.
"""

from ....logger import logger
from ...colors import Color
from ...graph import Arrow, Edge, Vertex
from ..filesystem.entities import File, Handle
from ..network.entities import Connection, Socket
from .entities import Data, Diff

__all__ = ["IsSimilarTo", "IsAlmostIn", "IsDiffOf", "IsReadBy", "WritesInto", "HasSimilarContent"]





logger.info("Loading relations from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

class IsSimilarTo(Arrow):

    """
    This kind of arrow indicates two Data nodes have a similar content. The direction of the arrow indicates the order of apparition.
    """

    label : str = ""

    source : Data
    destination : Data





class IsAlmostIn(Arrow):

    """
    This kind of arrow indicates that a Data node's content might be contained in another's one. The direction of the arrow indicates the order of apparition.
    """

    label : str = ""

    source : Data
    destination : Data





class IsDiffOf(Edge):

    """
    This kind of edge indicates that a Diff node sums up operations performed on the destination node.
    """

    label : str = ""

    source : File | Connection | Handle | Socket
    destination : Diff





class IsReadBy(Arrow):

    """
    This kind of arrow indicates that a Diff node only read from a vector.
    """

    source : Diff
    destination : File | Connection | Handle | Socket





class WritesInto(Arrow):

    """
    This kind of arrow indicates that a Diff node only wrote to a vector.
    """

    source : File | Connection | Handle | Socket
    destination : Diff





class HasSimilarContent(Edge):

    """
    This kind of arrow indicates that two Diff nodes have buffers with a certain similarity rate.
    """

    __slots__ = {
        "__source_buffer" : "The name of the buffer of the source node that the similarity was computed with.",
        "__destination_buffer" : "The name of the buffer of the destination node that the similarity was computed with."
    }

    __pickle_slots__ = {
        "source_buffer",
        "destination_buffer"
    }

    label : str = ""

    source : Diff
    destination : Diff

    def __init__(self, source: Vertex, destination: Vertex, *, c: Color | None = None, auto_write: bool = True) -> None:
        super().__init__(source, destination, c=c, auto_write=auto_write)
        self.__source_buffer = "global_buffer"
        self.__destination_buffer = "global_buffer"
        
    @property
    def source_buffer(self) -> str:
        """
        Returns the name of the content selected for comparison in the source node.
        """
        return self.__source_buffer
    
    @source_buffer.setter
    def source_buffer(self, name : str):
        if name not in ("read_buffer", "write_buffer", "global_buffer"):
            raise ValueError("Diff node buffers can only be set to one of ('read_buffer', 'write_buffer', 'global_buffer')")
        self.__source_buffer = name
    
    @property
    def destination_buffer(self) -> str:
        """
        Returns the name of the content selected for comparison in the destination node.
        """
        return self.__destination_buffer
    
    @destination_buffer.setter
    def destination_buffer(self, name : str):
        if name not in ("read_buffer", "write_buffer", "global_buffer"):
            raise ValueError("Diff node buffers can only be set to one of ('read_buffer', 'write_buffer', 'global_buffer')")
        self.__destination_buffer = name





del Arrow, Color, Connection, Data, Diff, Edge, File, Handle, Socket, Vertex, logger


"""
This module holds the CallHandler class of the execution behavioral package.
"""

from typing import Any, Callable

from Viper.meta.iterable import InstancePreservingClass

from ...utils import chrono
from .entities import Call


class CallHandler(metaclass = InstancePreservingClass):

    """
    The class for all handlers of system calls.
    Upon creation, a system call vertex will look into the instances of this class to find all valid handlers and execute its integration function.
    
    To make a handler, give an integration function (performs whatever you want with the Call vertex) and as many api names as you want to match them.
    """

    __sorted : dict[str, list["CallHandler"]] = {}
    __dynamic : dict[Callable[[str], bool], list["CallHandler"]] = {}
    __scanned : set[str] = set()

    @staticmethod
    def handlers(c : Call) -> list["CallHandler"]:
        """
        The list of all the CallHandlers to call in order to integrate this Call vertex.
        """
        if c.name not in CallHandler.__scanned:
            if c.name not in CallHandler.__sorted:
                CallHandler.__sorted[c.name] = []
            for d, l in CallHandler.__dynamic.items():
                if d(c.name):
                    CallHandler.__sorted[c.name].extend(l)
        return CallHandler.__sorted[c.name]
    
    @chrono
    @staticmethod
    def integrate_chain(c : Call):
        """
        Automatically calls all handlers responsible of this Call vertex.
        """
        for h in CallHandler.handlers(c):
            h.integrate(c)

    def __init__(self, integrator : Callable[[Call], None], *names : str | Callable[[str], bool]) -> None:
        if not callable(integrator) or any(not callable(n) and not isinstance(n, str) for n in names):
            types : list[str] = [type(integrator).__name]
            types.extend(type(n).__name__ for n in names)
            raise TypeError("Expected callable, one or more str, got " + ", ".join(types[:-1]) + " and " + types[-1])
        self.__names = names
        self.__integrator = integrator
        for n in self.__names:
            if isinstance(n, str):
                if n not in CallHandler.__sorted:
                    CallHandler.__sorted[n] = []
                CallHandler.__sorted[n].append(self)
            else:
                if n not in CallHandler.__dynamic:
                    CallHandler.__dynamic[n] = []
                CallHandler.__dynamic[n].append(self)
    

    def __repr__(self) -> str:
        """
        Implements repr(self).
        """
        return "CallHandler(" + repr(self.__integrator) + ", " + ", ".join(str(n) for n in self.__names) + ")"
    

    def match(self, c : Call) -> bool:
        """
        Returns True if given Call c should be handled by this CallHandler
        """
        if c.name in self.__names:
            return True
        return False
    

    def integrate(self, c : Call):
        """
        Integrates Call c into the graph
        """
        self.__integrator(c)
    

    def __contains__(self, c : Any) -> bool:
        """
        Implements c in self.
        Equivalent to self.match(c).
        """
        if not isinstance(c, Call):
            return False
        return self.match(c)
    
    
    def __call__(self, c : Call):
        """
        Implements self(c).
        Equivalent to self.integrate(c)
        """
        self.integrate(c)


"""
This package defines all the node types for building the graph. It contains multiple sub-packages that are behavioral models that BAGUETTE understands.
Each sub-package contains three modules : entities, relations and integration.
    - Entities modules contain all the Vertex subclasses defined for each behavioral model.
    - Relations modules contain all the Edge and Arrow subclasses defined for each behavioral model.
    - Integration modules contain special functions used by the compiler to integrate these entities and relations to BAGUETTES graphs.
Furthermore, for each sub-package, all entities and relations are also stored at package level.
"""

# Ground rules:
# - for each type package, define entities, relations and integration modules.
# - You should not reference relations or integrations globally in entities modules.
# - You should not reference integrations globally in relation modules.
# - Indeed these modules should be loaded in the order entities - relations - integrations globally (all entities first, then all relations, etc.)

from pathlib import Path
from importlib import import_module
modules = Path(__file__).parent.glob("*")

__all__ = [p.name for p in modules if p.is_dir() and "__init__.py" in [c.name for c in p.iterdir()]]

for m in __all__:
    import_module("." + m + ".entities", "baguette.bakery.source.types")
for m in __all__:
    import_module("." + m + ".relations", "baguette.bakery.source.types")
for m in __all__:
    import_module("." + m + ".integration", "baguette.bakery.source.types")
    del m

del Path, modules, import_module


"""
This module contains Vertex subclasses for this behavioral package.
"""

import pathlib

from ....logger import logger
from ...colors import Color
from ...graph import UniqueVertex, Vertex

__all__ = ["File", "Directory", "Handle"]





logger.info("Loading entities from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

class File(UniqueVertex):

    """
    A file vertex. Represents a (real) file opened during execution.
    """

    __slots__ = {
        "__path" : "The path to the file in the file system",
    }

    __pickle_slots__ = {
        "path"
    }

    def __init__(self, *, c: Color | None = None, parent: Vertex | None = None) -> None:
        from pathlib import PurePath
        super().__init__(c=c, parent=parent)
        self.__path : PurePath | None = None

    @property
    def path(self) -> pathlib.PurePath:
        """
        The absolute path to the file.
        """
        if self.__path is None:
            raise RuntimeError("Got a File without path.")
        return self.__path

    @path.setter
    def path(self, value : str):
        from ...utils import path_factory
        self.__path = path_factory(value)
    
    @property
    def name(self) -> str:
        """
        The name of the file (tail of the path).
        """
        if self.__path is None:
            raise RuntimeError("Got a File without path.")
        return self.__path.name
    
    @property
    def extension(self) -> str:
        """
        The file extention (lowercased without '.').
        """
        if self.__path is None:
            raise RuntimeError("Got a File without path.")
        return self.__path.suffix.lower().replace(".", "")
    
    @property
    def label(self) -> str:
        """
        Returns a label for this File node.
        """
        return 'File "' + self.name + '"'
    




class Directory(UniqueVertex):

    """
    A directory vertex. Represents a (real) directory opened during execution.
    """

    __slots__ = {
        "__path" : "The path to the directory in the file system",
    }

    __pickle_slots__ = {
        "path"
        }

    def __init__(self, *, c: Color | None = None, parent: Vertex | None = None) -> None:
        from pathlib import PurePath
        super().__init__(c=c, parent=parent)
        self.__path : PurePath | None = None

    @property
    def path(self) -> pathlib.PurePath:
        """
        The absolute path to the directory.
        """
        if self.__path is None:
            raise RuntimeError("Got a File without path.")
        return self.__path

    @path.setter
    def path(self, value : str):
        from ...utils import path_factory
        self.__path = path_factory(value)
    
    @property
    def name(self) -> str:
        """
        The name of the directory (tail of the path).
        """
        if self.__path is None:
            raise RuntimeError("Got a File without path.")
        return self.__path.name if self.__path.name else self.__path.drive
    
    @property
    def label(self) -> str:
        """
        Returns a label for this Directory node.
        """
        if self.__path is None:
            raise RuntimeError("Got a File without path.")
        if not self.__path.name:
            return 'Drive "' + self.name.replace(":", "") + '"'
        return 'Directory "' + self.name + '"'
    




class Handle(UniqueVertex):

    """
    A handle vertex. Represents a file handle, used when a program opens a file.
    """

    __slots__ = {
        # For all objects:
        "synchronize" : "If the handle can perform synchronization operations on the object",

        # For file objects:
        "read" : "If the handle has the right to read the file",
        "write" : "If the handle has the right to write to the file",
        "append" : "If the handle has the right to append to the file",
        "execute" : "If the handle has the right to execute the file",
        "read_attributes" : "If the handle has the right to read file attributes",
        "write_attributes" : "If the handle can change the file attributes",
        "read_extended_attributes" : "If the handle can read the extended file attributes",
        "write_extended_attributes" : "If the handle can change the extended file attributes",

        # For directory objects:
        "list_directory" : "If the handle has the right to list the items in the given directory",
        "traverse" : "If the handle has the right to traverse the directory and use it to access subforlders/files"
    }

    __pickle_slots__ = {
        "synchronize",
        "read",
        "write",
        "append",
        "execute",
        "read_attributes",
        "write_attributes",
        "read_extended_attributes",
        "write_extended_attributes",
        "list_directory",
        "traverse"
    }

    def __init__(self, *, c: Color | None = None, parent: Vertex | None = None) -> None:
        super().__init__(c=c, parent=parent)

        self.synchronize : bool = False

        self.read : bool = False
        self.write : bool = False
        self.append : bool = False
        self.execute : bool = False
        self.read_attributes : bool = False
        self.write_attributes : bool = False
        self.read_extended_attributes : bool = False
        self.write_extended_attributes : bool = False

        self.list_directory : bool = False
        self.traverse : bool = False
    
    @property
    def file(self) -> File | Directory:
        """
        Returns the file (or directory) node that this handle is working on.
        """
        from .relations import UsesDirectory, UsesFile
        for e in self.edges:
            if isinstance(e, UsesFile | UsesDirectory):
                return e.destination
        raise RuntimeError("Got a Handle working on no File or Directory.")






del Color, UniqueVertex, Vertex, logger, pathlib